"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const express_1 = __importDefault(require("express"));
const connect_timeout_1 = __importDefault(require("connect-timeout"));
const cors_1 = __importDefault(require("cors"));
const util_1 = __importDefault(require("util"));
const bridge_error_1 = __importDefault(require("./bridge-error"));
const outbound_response_1 = __importDefault(require("./outbound-response"));
class CoronadoBridge {
    /**
     * constructor
     * This function configures and starts CoronadoBridge.
     * 1. Instantiates the OutboundProvider
     * 2. Starts Express
     * @params config: IBridgeConfig
     * @returns - None
     **/
    constructor(config) {
        this.config = config;
        this.logger = config.logger;
        this.servers = [];
        this.corsOptions = config.corsOptions;
        this.requestTimeoutMs = config.requestTimeoutMs;
        // 1. Instantiates the OutboundProvider
        this.outboundProvider = config.outboundProvider;
        // 2. Starts Express
        this.setupExpress();
    }
    /**
     * setupExpress
     * This function creates express servers on the given ports.
     * 1. Create an express server for each port in the config
     * 2. Sets up a generic server route that takes in all requests
     * @returns - None
     **/
    setupExpress() {
        var _a, _b;
        (_a = this.logger) === null || _a === void 0 ? void 0 : _a.info(`CoronadoBridge - setupExpress - starting setup`);
        let ports = this.config.ports || [3000];
        // 1. Create an express server for each port in the config
        ports.forEach((port) => {
            var _a;
            let app = (0, express_1.default)();
            // Add body parsing middleware (the one used is determined by the
            // request content-type header.
            // Express Middleware are explained here: https://expressjs.com/en/4x/api.html#express.urlencoded
            // TODO: Some of the middleware choices might require passing additional
            //  options. These could be exposed in IBridgeConfig if needed. For now,
            //  these are all set to defaults per the docs.
            app.use(express_1.default.json(this.config.jsonParsingOptions)); // content-type: application/json
            app.use(express_1.default.text()); // content-type: text/plain
            // Notice we support the IBridgeConfig passing in cors options
            app.use((0, cors_1.default)(this.corsOptions));
            // 2. Sets up a generic server route
            app.all(
            // Takes all paths
            '*', 
            // Times out per the Coronado Config or defaults to 30s
            (0, connect_timeout_1.default)(((_a = this.requestTimeoutMs) === null || _a === void 0 ? void 0 : _a.toString()) || '30s'), 
            // Defines our handler
            (req, res, next) => {
                var _a, _b;
                // Disregard HTTP requests that are *not* GET, POST, PUT, DELETE
                if (['GET', 'POST', 'PUT', 'DELETE'].indexOf(req.method) < 0) {
                    (_a = this.logger) === null || _a === void 0 ? void 0 : _a.info(`CoronadoBridge - request received with method ${req.method} -> will not pass to provider`);
                    // OPTION still needs to be handled by cors. So we have to give this
                    // to next() which is really the right way to handle this anyway
                    next();
                    return;
                }
                (_b = this.logger) === null || _b === void 0 ? void 0 : _b.info(`CoronadoBridge - request received on port:${port} -> passing message to outboundProvider`);
                // Merge request body/query into one object for the message and add any passed params.
                // Note: Query properties override body properties (example URL: /?exchange=test&topic=coolfactor1)
                let providerReq = {
                    method: req.method,
                    body: req.body,
                    query: req.query,
                    headers: req.headers,
                    // params will be set below...
                };
                if (req.params) {
                    // This is basically parsing out what comes in like this:
                    //   "params": {
                    //     "0": "/param1/param2/param3/param4"
                    //   }
                    // Into:
                    //   "params": [
                    //     "param1",
                    //     "param2",
                    //     "param3",
                    //     "param4"
                    //   ]
                    // @ts-ignore
                    const paramsArray = req.params['0'].split('/');
                    paramsArray.shift(); // The first item is always undefined so lets remove it.
                    // Now set it to our object
                    providerReq.params = paramsArray;
                }
                this.outboundProvider
                    .handler(providerReq)
                    .then((handlerResponse) => {
                    var _a, _b, _c, _d, _e;
                    (_a = this.logger) === null || _a === void 0 ? void 0 : _a.info(`CoronadoBridge - handler processed message`); // Inspect the received handlerResponse to see if we have the structured class
                    // OutboundResponse. If we do, we treat that special.
                    if (handlerResponse instanceof outbound_response_1.default) {
                        (_b = this.logger) === null || _b === void 0 ? void 0 : _b.debug(`CoronadoBridge - received an OutboundResponse <IProviderRes>. Will respond per that object.`);
                        (_c = this.logger) === null || _c === void 0 ? void 0 : _c.trace(`CoronadoBridge - OutboundResponse: ${util_1.default.inspect(handlerResponse)}`);
                        if (handlerResponse.headers) {
                            const headerKeys = Object.keys(handlerResponse.headers);
                            headerKeys.forEach((key, index) => {
                                // @ts-ignore
                                res.header(key, handlerResponse.headers[key]);
                            });
                        }
                        res.status(handlerResponse.status).send(handlerResponse.body);
                    }
                    else {
                        // it's not our ObjectResponse, so just default to a simple output
                        (_d = this.logger) === null || _d === void 0 ? void 0 : _d.debug(`CoronadoBridge - received a generic object. Auto-responding with HTTP=200 and the object.`);
                        (_e = this.logger) === null || _e === void 0 ? void 0 : _e.info(`CoronadoBridge - Check your HTTP CLIENT for the response!`);
                        res.status(200).send(handlerResponse);
                    }
                })
                    .catch((error) => {
                    var _a, _b;
                    if (error instanceof bridge_error_1.default) {
                        (_a = this.logger) === null || _a === void 0 ? void 0 : _a.error(`CoronadoBridge - handler error ${error.status} - ${error.message}`);
                        res.status(error.status).send({
                            // Just send out the error object
                            error: {
                                status: error.status,
                                message: error.message,
                                name: error.name,
                            },
                        });
                    }
                    else {
                        (_b = this.logger) === null || _b === void 0 ? void 0 : _b.error(`CoronadoBridge - handler error ${error}`);
                        res.status(500).send({
                            message: error,
                        });
                    }
                });
            });
            // Setup a default error handler (this should be LAST in the app.use chain!)
            app.use((err, req, res, next) => {
                var _a;
                // On error, check to see what we can do
                (_a = this.logger) === null || _a === void 0 ? void 0 : _a.error(`CoronadoBridge Error: ${err.message}`);
                // Check for some expected errors
                if (err.code === 'ETIMEDOUT') {
                    // It's a timeout, so respond with client timeout
                    res.status(408).send(err.message);
                }
                else {
                    // It's something else, so we return server-error and the verbose message
                    res.status(500).send(err.message);
                }
            });
            // With all the express setup one, let's tart the server
            const server = app.listen(port, () => {
                var _a;
                (_a = this.logger) === null || _a === void 0 ? void 0 : _a.info(`CoronadoBridge - started express server on ${port}`);
            });
            this.servers.push(server);
        });
        // Log
        (_b = this.logger) === null || _b === void 0 ? void 0 : _b.info(`CoronadoBridge - setupExpress - setup complete`);
    }
    close() {
        this.servers.forEach((server, index) => {
            server.close(() => {
                var _a;
                (_a = this.logger) === null || _a === void 0 ? void 0 : _a.info(`CoronadoBridge - Closed server ${index}`);
            });
        });
    }
}
exports.default = CoronadoBridge;
